import { BaseSchema, Output } from 'valibot';
import { z } from 'zod';

interface Adapter {
    /**
     * @returns Whether we are in the browser
     */
    isBrowser(): boolean;
    /**
     * Note: The first param includes the `?` prefix
     */
    updateServerUrl: QueryUpdater;
    /**
     * Note: The first param includes the `?` prefix
     */
    updateBrowserUrl: QueryUpdater;
    /**
     * A custom function to retrieve the URL when in the browser
     *
     * @default () => {@linkcode window.location}
     */
    getBrowserUrl: QueryFetcher;
    /**
     * A custom function to retrieve the URL when on the server
     */
    getServerUrl: QueryFetcher;
}

/// <reference lib="dom" />

/** Use this over valibot AnySchema type as we need to widen the type of `schema` to be any string */
type AnySchema<TOutput = any> = BaseSchema<any, TOutput> & {
    schema: string;
};
type ValibotValidator = AnySchema;
type FunctionValidator<TOut = unknown> = (value?: string) => TOut;
type ZodValidator = z.ZodType;
type Validator = FunctionValidator | ZodValidator | ValibotValidator;
type inferFromValidator<TValidator extends Validator> = TValidator extends ZodValidator ? z.infer<TValidator> : TValidator extends ValibotValidator ? Output<TValidator> : TValidator extends FunctionValidator ? ReturnType<TValidator> : never;
type QuerySchema = Record<string, Validator>;
type inferShape<TShape extends QuerySchema> = {
    [K in keyof TShape]: inferFromValidator<TShape[K]>;
} & {};
/**
 * @param search Includes the `?` prefix
 */
type QueryUpdater = (search: string, hash: string) => void;
type QueryFetcher = () => {
    search: string;
    hash: string;
};
type Serializer = (value: unknown) => string;
interface QueryParamsOptions {
    /**
     * Provide a custom implementation of {@linkcode window}. It requires
     * {@linkcode window.location}, {@linkcode window.history},
     * {@linkcode window.addEventListener} and {@linkcode window.removeEventListener}
     *
     * @default window
     */
    windowObj?: Pick<typeof window, "location" | "history" | "addEventListener" | "removeEventListener">;
    /**
     * Add a delay (in ms) before updating the browser URL. This is useful in
     * situations where URL updates happen frequently, e.g., on every keystroke.
     *
     * Note this does not affect the query params rune - this will always be
     * updated optimistically.
     *
     * @default 0
     */
    debounce?: number;
    /**
     * Control how query params are serialised to the browser query params
     *
     * **Note**: this is NOT for encoding values into URI components - it is
     * for serialising values into strings, which will then be encoded
     * internally.
     */
    serialise?: Serializer;
    adapter?: Adapter;
}
type QueryParams<TShape extends Record<string, unknown>> = TShape & Params<TShape>;
interface Params<TShape extends object> {
    /**
     * The raw query params, parsed from {@linkcode windowObj.location.href}
     *
     * Note: this may include query params not defined in your schema
     */
    readonly raw: Record<string, string>;
    /**
     * The unmodified query params parsed from the {@linkcode raw} params
     *
     * Note: this may include query params not defined in your schema and will be
     * passed through as-is (as strings)
     */
    readonly query: TShape;
    /**
     * The query string, generated from the {@linkcode query} which may contain
     * query params not defined in your schema.
     */
    readonly search: string;
    /** Replace _ALL_ query params, triggering a reactive and browser update */
    set(params: TShape): void;
    /** Update a subset of the query params, triggering a reactive and browser update */
    update(params: Partial<TShape>): void;
    /** Remove query params. Note that this may cause your validation check to fail */
    remove(...params: (keyof TShape)[]): void;
    /** Manually unset unregister all event listeners */
    unsubscribe(): void;
    /** Return the query keys. Unlike {@linkcode Object.keys}, this is type-safe */
    keys(): (keyof TShape)[];
    entries(): [keyof TShape, TShape[keyof TShape]][];
    [Symbol.dispose]: () => void;
}

export { AnySchema as A, FunctionValidator as F, Params as P, QuerySchema as Q, Serializer as S, ValibotValidator as V, ZodValidator as Z, QueryParamsOptions as a, QueryParams as b, Validator as c, inferFromValidator as d, QueryUpdater as e, QueryFetcher as f, Adapter as g, inferShape as i };
